// interface to the aircraft
// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// http://creativecommons.org/licenses/by-nc-sa/3.0/   
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// $Id: aircraft.c 400 2011-10-21 23:10:51Z andy $

#include "ppm.h"
#include "aircraft.h"
#include "configuration.h"

// equivalent to a single click of trim sliders on regular transmitter
// a percentage change in the trim setting
#define TRIM 3

// length of pulse when stick and trim are in center position, in microseconds
#define CENTERPOSITION 1120
// change in pulse length from center position at 100% stick throw volume
#define STICKFULLTHROW 420
// change in pulse length from center position at 100% trim throw volume
#define TRIMFULLTHROW 120

// holds the current stick and trim values for each channel as
// a percentage -100 to 100 with zero = center
static int StickPosition[MAXCHANNELS];
static int TrimPosition[MAXCHANNELS];
// holds flags to indicate which channels are reversed
// true = reversed, false = not reversed
static int Reversed[MAXCHANNELS];
// holds the current travel volumes for each channel as
// a percentage 0 to 100
static int LeftDownTravelVolume[MAXCHANNELS];
static int RightUpTravelVolume[MAXCHANNELS];
// the number of channels being used
static int NumChannels;

// module level functions
static void Aircraft_UpdateChannel
    (
    int Channel                                            // channel to update
    );

// initializes the module
// starts communication with the aircraft via the transmitter
void Aircraft_Init
  (
  int NumberofChannels                                     // number of channels
  )
{
    int Channel;

    // remember the number of channels
    NumChannels = NumberofChannels;

    // make sure PPM is stopped
    PPM_Stop();

    // center stick and trim positions
    // not reversed
    // 100% travel volume in all directions
    for (Channel = 0; Channel < NumChannels; Channel++)
    {
        StickPosition[Channel] = 0;
        TrimPosition[Channel] = 0;
        Reversed[Channel] = 0;
        LeftDownTravelVolume[Channel] = 100;
        RightUpTravelVolume[Channel] = 100;
    }

    // transmit all channels in center position
    PPM_Init(CENTERPOSITION, NumChannels);

    // start communicating with aircraft   
    PPM_Start();
}

// sends the current stick and trim values to the aircraft
static void Aircraft_UpdateChannel
    (
    int Channel                                            // channel to update (1 - NumChannels)
    )
{
    // converts stick and trim positions from percentage to microseconds
    // using: microseconds = percentage / 100 * fullthrow * travelvolume / 100

    int StickTravelVolume;
    int TrimTravelVolume;

    // ignore invalid channels
    if ((Channel < 1) || (Channel > NumChannels)) return;

    // if stick position is positive then use travel volume for the position directions
    // otherwise use the travel volume for the negative directions
    if (StickPosition[Channel - 1] >= 0)
        StickTravelVolume = RightUpTravelVolume[Channel - 1];
    else
        StickTravelVolume = LeftDownTravelVolume[Channel - 1];

    // if the trim position is positive then use travel volume for the position directions
    // otherwise use the travel volume for the negative directions
    if (TrimPosition[Channel - 1] >= 0)
        TrimTravelVolume = RightUpTravelVolume[Channel - 1];
    else
        TrimTravelVolume = LeftDownTravelVolume[Channel - 1];

    if (Reversed[Channel - 1])
    {
        PPM_SetChannelValue(
            Channel - 1,
            CENTERPOSITION -
            ((StickPosition[Channel - 1] * STICKFULLTHROW * StickTravelVolume) / 10000) -
            ((TrimPosition[Channel - 1] * TRIMFULLTHROW * TrimTravelVolume) / 10000)
        );
    }
    else
    {
        PPM_SetChannelValue(
            Channel - 1,
            CENTERPOSITION +
            ((StickPosition[Channel - 1] * STICKFULLTHROW * StickTravelVolume) / 10000) +
            ((TrimPosition[Channel - 1] * TRIMFULLTHROW * TrimTravelVolume) / 10000)
        );
    }
}

// gets the stick value for a channel
int Aircraft_GetStick
  (
  int Channel                                              // channel to get (1 - NumChannels)
  )
{
  // ignore invalid channels
  if ((Channel < 1) || (Channel > NumChannels)) return 0;

  return StickPosition[Channel - 1];
}

// sets the stick value for a channel
void Aircraft_SetStick
    (
    int Channel,                                           // channel to set (1 - NumChannels)
    int Value                                              // percentage -100 to 100 (0 = center)
    )
{
    // ignore invalid channels
    if ((Channel < 1) || (Channel > NumChannels)) return;

    // make sure value is valid
    if (Value < -100) Value = -100;
    if (Value > 100) Value = 100;

    // store
    StickPosition[Channel - 1] = Value;

    // send to aircraft
    Aircraft_UpdateChannel(Channel);
}

// sets the trim value for a channel
void Aircraft_SetTrim
    (
    int Channel,                                           // channel to set (1 - NumChannels)
    int Value                                              // percentage -100 to 100 (0 = center)
    )
{
    // ignore invalid channels
    if ((Channel < 1) || (Channel > NumChannels)) return;

    // make sure value is valid
    if (Value < -100) Value = -100;
    if (Value > 100) Value = 100;
  
    // store
    TrimPosition[Channel - 1] = Value;

    // send to aircraft
    Aircraft_UpdateChannel(Channel);
}

// gets the trim value for a channel
int Aircraft_GetTrim
  (
  int Channel                                              // channel to set (1 - NumChannels)
  )
{
  // ignore invalid channels
  if ((Channel < 1) || (Channel > NumChannels)) return 0;

  return TrimPosition[Channel - 1];
}

// sets the travel volume for a channel
void Aircraft_SetTravelVolume
    (
    int Channel,                                           // channel to set (1 - NumChannels)
    direction Direction,                                   // direction of channel volume
    int Value                                              // percentage 10 to 120
    )
{
    // ignore invalid channels
    if ((Channel < 1) || (Channel > NumChannels)) return;

    // make sure value is valid
    if (Value < 10) Value = 10;
    if (Value > 120) Value = 120;
  
    // store
    if (Direction == LeftDown)
        LeftDownTravelVolume[Channel - 1] = Value;
    else
        RightUpTravelVolume[Channel - 1] = Value;
    
    // send to aircraft
    Aircraft_UpdateChannel(Channel);
}

// increases the trim on a channel
// returns new trim position
int Aircraft_AddTrimByAmount
    (
    int Channel,                                           // channel to set (1 - NumChannels)
    int Amount                                             // amount to increase trim by in percentage
    )
{
    // ignore invalid channels
    if ((Channel < 1) || (Channel > NumChannels)) return TrimPosition[Channel - 1];

    Aircraft_SetTrim(Channel, TrimPosition[Channel - 1] + Amount);

    return TrimPosition[Channel - 1];
}

// increases the trim on a channel
// equivalent to moving the trim slider one click on a regular transmitter
// returns new trim value
int Aircraft_AddTrim
    (
    int Channel                                            // channel to set (1 - NumChannels)
    )
{
  return Aircraft_AddTrimByAmount(Channel, TRIM);
}

// decreases the trim on a channel
// returns new trim value
int Aircraft_SubtractTrimByAmount
    (
    int Channel,                                           // channel to set (1 - NumChannels)
    int Amount                                             // amount to decreate trim by in percentage
    )
{
    // ignore invalid channels
    if ((Channel < 1) || (Channel > NumChannels)) return TrimPosition[Channel - 1];

    Aircraft_SetTrim(Channel, TrimPosition[Channel - 1] - Amount);

    return TrimPosition[Channel - 1];
}

// dereases the trim on a channel
// equivalent to moving the trim slider one click on a regular transmitter
// returns new trim value
int Aircraft_SubtractTrim
    (
    int Channel                                            // channel to set (1 - NumChannels)
    )
{
  return Aircraft_SubtractTrimByAmount(Channel, TRIM);
}

// sets the servo direction on a channel
void Aircraft_SetChannelDirection
    (
    int Channel,                                           // channel to configure (1 - NumChannels)
    int Reverse                                            // zero = not reversed, non-zero = reversed
    )
{
    // ignore invalid channels
    if ((Channel < 1) || (Channel > NumChannels)) return;

    // mark the channel direction
    Reversed[Channel - 1] = Reverse;

    // send current values to aircraft
    Aircraft_UpdateChannel(Channel);
}
