// command parser
// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// http://creativecommons.org/licenses/by-nc-sa/3.0/
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// $Id: commandparser.c 399 2011-10-20 21:51:15Z andy $

/*
Commands:
Commands are sent and received in ASCII.
All values sent and received are an ASCII representation of decimal integers.

<command> <direction> [<parameter] [<parameter>] ... CR+LF

Command is two letters.
Direction is S for set or G for get.
The number of parameters and their contents vary depending on the command.
Commands and responses are terminated with a carriage return and a line feed.
Set commands return "ERROR" or "OK".
Get commands return a value or "ERROR".
Invalid parameters return "ERROR" for the command.
Commands don't return until the operation has completed.

SE S aircraftnum
  Select aircraft (1 - 7)
  Applied immediately

SE G
  Get the currently selected aircraft (1 - 7)

CH S aircraftnum channelnum
  Set the number of channels (1 - 8)
  Requires a reset to be applied

CH G aircraftnum
  Get the number of channels (1 - 8)

NA S aircraftnum name
  Set the name of an aircraft (max 31 ASCII characters)

NA G aircraftnum
  Get the name of an aircraft (max 31 ASCII characters)

RE S aircraftnum channelnum state
  Sets the channel reversing. If state = 1 then reversed, 0 = not reversed
  If changing current aircraft then the change is applied immediately

RE G aircraftnum channelnum
  Gets the channel reversing. Returns 1 = reversed, 0 = not reversed

TR S aircraftnum channelnum
  Sets the channel trim value (-100 - 100)
  If changing current aircraft then the change is applied immediately

TR G aircraftnum channelnum
  Gets the channel trim value (-100 - 100)

LD S aircraftnum channelnum
  Sets the left/down travel volume (10 - 120)
  If changing current aircraft then the change is applied immediately

LD G aircraftnum channelnum
  Gets the left/down travel volume (10 - 120)

RU S aircraftnum channelnum
  Sets the right/up travel volume (10 - 120)
  If changing current aircraft then the change is applied immediately

RU G aircraftnum channelnum
  Gets the right/up travel volume (10 - 120)

PO S channelnum postion
  Sets the channel stick position (-100 - 100)
  Applied immediately

PO G channelnum
  Gets the channel stick position (-100 - 100)

CA S aircraftnum throttlechannelnum aileronschannelnum rudderchannelnum elevatorschannelnum
  Sets the channel assignments for the aircraft
  Requires a reset to be applied

CA G aircraftnum
  Gets the channel assignments for the aircraft Returned as a space separated list:
  throttlechannelnum aileronschannelnum rudderchannelnum elevatorschannelnum

AS S
  Apply current settings

*/

#include "commandparser.h"
#include "serial.h"
#include "library.h"
#include "configuration.h"
#include "aircraft.h"
#include "core.h"

// command handler function pointer type
typedef void (*commandhandler_t)(char *Parameters, int ParametersLength);

typedef struct
{
  char *CommandString;
  commandhandler_t Handler;
} commandtableentry_t;

// command handlers
static void CommandParser_SetAircraft(char *Parameters, int ParametersLength);
static void CommandParser_GetAircraft(char *Parameters, int ParametersLength);
static void CommandParser_SetNumberofChannels(char *Parameters, int ParametersLength);
static void CommandParser_GetNumberofChannels(char *Parameters, int ParametersLength);
static void CommandParser_SetAircraftName(char *Parameters, int ParametersLength);
static void CommandParser_GetAircraftName(char *Parameters, int ParametersLength);
static void CommandParser_SetChannelReverse(char *Parameters, int ParametersLength);
static void CommandParser_GetChannelReverse(char *Parameters, int ParametersLength);
static void CommandParser_SetChannelTrim(char *Parameters, int ParametersLength);
static void CommandParser_GetChannelTrim(char *Parameters, int ParametersLength);
static void CommandParser_SetChannelLeftDownTravelVolume(char *Parameters, int ParametersLength);
static void CommandParser_GetChannelLeftDownTravelVolume(char *Parameters, int ParametersLength);
static void CommandParser_SetChannelRightUpTravelVolume(char *Parameters, int ParametersLength);
static void CommandParser_GetChannelRightUpTravelVolume(char *Parameters, int ParametersLength);
static void CommandParser_SetChannelPosition(char *Parameters, int ParametersLength);
static void CommandParser_GetChannelPosition(char *Parameters, int ParametersLength);
static void CommandParser_SetChannelAssignments(char *Parameters, int ParametersLength);
static void CommandParser_GetChannelAssignments(char *Parameters, int ParametersLength);
static void CommandParser_ApplySettings(char *Parameters, int ParametersLength);

// supported commands
static commandtableentry_t Commands[] = {
  {"SE S", CommandParser_SetAircraft},
  {"SE G", CommandParser_GetAircraft},
  {"CH S", CommandParser_SetNumberofChannels},
  {"CH G", CommandParser_GetNumberofChannels},
  {"NA S", CommandParser_SetAircraftName},
  {"NA G", CommandParser_GetAircraftName},
  {"RE S", CommandParser_SetChannelReverse},
  {"RE G", CommandParser_GetChannelReverse},
  {"TR S", CommandParser_SetChannelTrim},
  {"TR G", CommandParser_GetChannelTrim},
  {"LD S", CommandParser_SetChannelLeftDownTravelVolume},
  {"LD G", CommandParser_GetChannelLeftDownTravelVolume},
  {"RU S", CommandParser_SetChannelRightUpTravelVolume},
  {"RU G", CommandParser_GetChannelRightUpTravelVolume},
  {"PO S", CommandParser_SetChannelPosition},
  {"PO G", CommandParser_GetChannelPosition},
  {"CA S", CommandParser_SetChannelAssignments},
  {"CA G", CommandParser_GetChannelAssignments},
  {"AS S", CommandParser_ApplySettings}
};

// buffer to receive serial commands
static char Command[SERIAL_MAXCOMMANDLENGTH];
// number of supported commands;
static int TotalCommands;

// initializes command parser module
void CommandParser_Init
  (
  void
  )
{
  // get number of supported commands
  TotalCommands = sizeof(Commands) / sizeof(commandtableentry_t);

  Serial_Init();
}

// gets a pointer to the next parameter or null if not found
static char *CommandParser_GetNextParameter
  (
  char *Parameters
  )
{
  char *NextParameter = NULL;
  int c;

  for (c = 0; c < strlen(Parameters); c++)
  {
    if (Parameters[c] == ' ')
    {
      NextParameter = Parameters + c + 1;
      return NextParameter;
    }
  }

  return NULL;
}

// parses the next command received, if any
void CommandParser_Parse
  (
  void
  )
{
  int32_t CommandLength;
  int i;
  int c;

  // if no command waiting then nothing to do
  if ((CommandLength = Serial_GetCommand(Command)) == 0) return;

  // ignore commands that are too short
  if (CommandLength < 4)
  {
    Serial_SendResponse(Error, False);
    return;
  }
  
  // third character is always a space
  if (Command[2] != ' ')
  {
    Serial_SendResponse(Error, False);
    return;
  }
    
  // convert to uppercase
  for (i = 0; i < CommandLength; i++)
  {
    if (Command[i] >= 'a' && Command[i] <= 'z') Command[i] -= 'a' - 'A';
  }  

  Core_FeedWatchdog();

  // search for a command handler
  for (c = 0; c < TotalCommands; c++)
  {
    // if found then call handler
    if (Command[0] == Commands[c].CommandString[0] &&
        Command[1] == Commands[c].CommandString[1] &&
        Command[3] == Commands[c].CommandString[3])
    {
      Commands[c].Handler(Command + 4, CommandLength - 4);
      return;
    }
  }

  Core_FeedWatchdog();

  // unknown command
  Serial_SendResponse(Error, False);
}

// handles get aircraft command
// gets the current aircraft
static void CommandParser_GetAircraft
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  char Buffer[10];

  AircraftNumber = Configuration_GetCurrentAircraft();

  Core_FeedWatchdog();

  Serial_SendString(itoa(AircraftNumber, Buffer, 10));
}

// handles set aircraft command
// sets the current aircraft
static void CommandParser_SetAircraft
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;  

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // parse
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // store
  Configuration_SetCurrentAircraft(AircraftNumber);

  Serial_SendResponse(Ok, True);

  // reset the microcontroller to use the new aircraft
  Core_Reset();
}

// handles set number of channels comamnd
static void CommandParser_SetNumberofChannels
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t NumberofChannels;  

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  NumberofChannels = atoi(Parameters);
  if (NumberofChannels < 1 || NumberofChannels > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  Configuration_SetAircraftNumChannels(AircraftNumber, NumberofChannels);

  Serial_SendResponse(Ok, False);
}

// handles get number of channels comamnd
static void CommandParser_GetNumberofChannels
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t NumberofChannels;
  char Buffer[10];  

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get the current setting
  NumberofChannels = Configuration_GetAircraftNumChannels(AircraftNumber);

  Core_FeedWatchdog();

  // return value
  Serial_SendString(itoa(NumberofChannels, Buffer, 10));
}

// handles set aircraft name comamnd
static void CommandParser_SetAircraftName
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // store
  Configuration_SetAircraftName(AircraftNumber, Parameters);

  Serial_SendResponse(Ok, False);
}

// handles get aircraft name comamnd
static void CommandParser_GetAircraftName
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  char Name[MAXAIRCRAFTNAME];

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // read name
  Configuration_GetAircraftName(AircraftNumber, Name);

  // return
  Serial_SendString(Name);
}

// handles set channel reverse comamnd
static void CommandParser_SetChannelReverse
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  int State;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel state
  State = atoi(Parameters);
  if (State != 0 && State != 1)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // store
  Configuration_SetChannelDirection(AircraftNumber, ChannelNumber, State);

  // if configuring current aircraft then apply change immediately
  if (AircraftNumber == Configuration_GetCurrentAircraft())
  {
    Aircraft_SetChannelDirection(ChannelNumber, State);
  }

  Serial_SendResponse(Ok, False);
}

// handles get channel reverse command
static void CommandParser_GetChannelReverse
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  int State;
  char Buffer[10];

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // get channel direction
  State = Configuration_GetChannelDirection(AircraftNumber, ChannelNumber);

  Serial_SendString(itoa(State, Buffer, 10));
}

// handles set channel trim comamnd
static void CommandParser_SetChannelTrim
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  int32_t Trim;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel trim
  Trim = atoi(Parameters);
  if (Trim < -100) Trim = -100;
  if (Trim >  100) Trim =  100;

  Core_FeedWatchdog();

  // store
  Configuration_SetChannelTrim(AircraftNumber, ChannelNumber, Trim);

  // if configuring current aircraft then apply change immediately
  if (AircraftNumber == Configuration_GetCurrentAircraft())
  {
    Aircraft_SetTrim(ChannelNumber, Trim);
  }

  Serial_SendResponse(Ok, False);
}

// handles get channel trim command
static void CommandParser_GetChannelTrim
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  int32_t Trim;
  char Buffer[10];

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // get
  Trim = Configuration_GetChannelTrim(AircraftNumber, ChannelNumber);

  Serial_SendString(itoa(Trim, Buffer, 10));
}

// handles set channel left/down travel volume comamnd
static void CommandParser_SetChannelLeftDownTravelVolume
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  uint32_t TravelVolume;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get travel volume
  TravelVolume = atoi(Parameters);
  if (TravelVolume < 20)  TravelVolume = 20;
  if (TravelVolume > 120) TravelVolume = 120;

  Core_FeedWatchdog();

  // store
  Configuration_SetChannelLeftDownTravelVolume(AircraftNumber, ChannelNumber, TravelVolume);

  // if configuring current aircraft then apply change immediately
  if (AircraftNumber == Configuration_GetCurrentAircraft())
  {
    Aircraft_SetTravelVolume(ChannelNumber, LeftDown, TravelVolume);
  }

  Serial_SendResponse(Ok, False);
}

// handles get channel left/down travel volume command
static void CommandParser_GetChannelLeftDownTravelVolume
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  uint32_t TravelVolume;
  char Buffer[10];

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // get
  TravelVolume = Configuration_GetChannelLeftDownTravelVolume(AircraftNumber, ChannelNumber);

  Serial_SendString(itoa(TravelVolume, Buffer, 10));
}

// handles set channel right/up travel volume comamnd
static void CommandParser_SetChannelRightUpTravelVolume
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  uint32_t TravelVolume;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get travel volume
  TravelVolume = atoi(Parameters);
  if (TravelVolume < 20)  TravelVolume = 20;
  if (TravelVolume > 120) TravelVolume = 120;

  Core_FeedWatchdog();

  // store
  Configuration_SetChannelRightUpTravelVolume(AircraftNumber, ChannelNumber, TravelVolume);

  // if configuring current aircraft then apply change immediately
  if (AircraftNumber == Configuration_GetCurrentAircraft())
  {
    Aircraft_SetTravelVolume(ChannelNumber, RightUp, TravelVolume);
  }

  Serial_SendResponse(Ok, False);
}

// handles get channel right/up travel volume command
static void CommandParser_GetChannelRightUpTravelVolume
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ChannelNumber;
  uint32_t TravelVolume;
  char Buffer[10];

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // get
  TravelVolume = Configuration_GetChannelRightUpTravelVolume(AircraftNumber, ChannelNumber);

  Serial_SendString(itoa(TravelVolume, Buffer, 10));
}

// handles set channel position comamnd
static void CommandParser_SetChannelPosition
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t ChannelNumber;
  int32_t Position;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get position
  Position = atoi(Parameters);
  if (Position < -100) Position = -100;
  if (Position >  100) Position =  100;

  Core_FeedWatchdog();

  // set value
  Aircraft_SetStick(ChannelNumber, Position);

  Serial_SendResponse(Ok, False);
}

// handles get channel position command
static void CommandParser_GetChannelPosition
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t ChannelNumber;
  int32_t Position;
  char Buffer[10];

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get channel number
  ChannelNumber = atoi(Parameters);
  if (ChannelNumber < 1 || ChannelNumber > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // get value
  Position = Aircraft_GetStick(ChannelNumber);

  Serial_SendString(itoa(Position, Buffer, 10));
}

// handles set channel assignments command
static void CommandParser_SetChannelAssignments
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ThrottleChannel, AileronsChannel, RudderChannel, ElevatorsChannel;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }
  // get throttle channel number
  ThrottleChannel = atoi(Parameters);
  if (ThrottleChannel < 1 || ThrottleChannel > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }
  // get ailerons channel number
  AileronsChannel = atoi(Parameters);
  if (AileronsChannel < 1 || AileronsChannel > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }
  // get rudder channel number
  RudderChannel = atoi(Parameters);
  if (RudderChannel < 1 || RudderChannel > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get next parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }
  // get elevators channel number
  ElevatorsChannel = atoi(Parameters);
  if (ElevatorsChannel < 1 || ElevatorsChannel > MAXCHANNELS)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // store
  Configuration_SetChannelAssignments(AircraftNumber, ThrottleChannel, AileronsChannel, RudderChannel, ElevatorsChannel);

  Serial_SendResponse(Ok, False);
}

// handles get channel assignments command
static void CommandParser_GetChannelAssignments
  (
  char *Parameters,
  int ParametersLength
  )
{
  uint32_t AircraftNumber;
  uint32_t ThrottleChannel, AileronsChannel, RudderChannel, ElevatorsChannel;
  char ThrottleBuffer[10], AileronsBuffer[10], RudderBuffer[10], ElevatorsBuffer[10];
  char ResultBuffer[40];
  char *ptr;
  int c;

  // get first parameter
  Parameters = CommandParser_GetNextParameter(Parameters);
  if (Parameters == NULL)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  // get aircraft number
  AircraftNumber = atoi(Parameters);
  if (AircraftNumber < 1 || AircraftNumber > MAXAIRCRAFT)
  {
    Serial_SendResponse(Error, False);
    return;
  }

  Core_FeedWatchdog();

  // get
  Configuration_GetChannelAssignments(AircraftNumber, &ThrottleChannel, &AileronsChannel, &RudderChannel, &ElevatorsChannel);

  itoa(ThrottleChannel, ThrottleBuffer, 10);
  itoa(AileronsChannel, AileronsBuffer, 10);
  itoa(RudderChannel, RudderBuffer, 10);
  itoa(ElevatorsChannel, ElevatorsBuffer, 10);
  
  Core_FeedWatchdog();

  ptr = ResultBuffer;

  for (c = 0; c < strlen(ThrottleBuffer); c++)
  {
    *ptr++ = ThrottleBuffer[c];
  }
  *ptr++ = ' ';

  for (c = 0; c < strlen(AileronsBuffer); c++)
  {
    *ptr++ = AileronsBuffer[c];
  }
  *ptr++ = ' ';

  for (c = 0; c < strlen(RudderBuffer); c++)
  {
    *ptr++ = RudderBuffer[c];
  }
  *ptr++ = ' ';

  for (c = 0; c < strlen(ElevatorsBuffer); c++)
  {
    *ptr++ = ElevatorsBuffer[c];
  }
  *ptr++ = NULL;

  Serial_SendString(ResultBuffer);
}

// handles apply settings command
static void CommandParser_ApplySettings
  (
  char *Parameters,
  int ParametersLength
  )
{
  Serial_SendResponse(Ok, True);

  // reset microcontroller
  Core_Reset();
}
