// configuration module
// stores configuration data in non-volatile memory
// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// http://creativecommons.org/licenses/by-nc-sa/3.0/
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// $Id: configuration.c 399 2011-10-20 21:51:15Z andy $

#include "configuration.h"
#include "i2c.h"
#include "core.h"

/*
Memory Map for 8kb EEPROM

0x0000 - 0x03FF: Global configuration
0x0400 - 0x07FF: Aircraft 1 record
0x0800 - 0x0BFF: Aircraft 2 record
0x0C00 - 0x0FFF: Aircraft 3 record
0x1000 - 0x03FF: Aircraft 4 record
0x1400 - 0x17FF: Aircraft 5 record
0x1800 - 0x1BFF: Aircraft 6 record
0x1C00 - 0x1FFF: Aircraft 7 record

Global configuration (1024 bytes):
0x0000 - 0x0003: Current aircraft number
0x0004 - 0x0007: Last known throttle value
0x0008 - 0x03FF: Reserved

Aircraft record (addresses are offset from base of record) (1024 bytes):
0x0000 - 0x0003: Number of channels
0x0004 - 0x001F: Reserved
0x0020 - 0x003F: Aircraft name (ASCII, null terminated)
0x0040 - 0x007F: Channel 1 record
0x0080 - 0x00BF: Channel 2 record
0x0200 - 0x023F: Channel 8 record
0x0240 - 0x0243: Throttle channel number
0x0244 - 0x0247: Ailerons channel number
0x0248 - 0x024B: Rudder channel number
0x024C - 0x024F: Elevators channel number
0x0250 - 0x03FF: Reserved

Channel Record (addresses are offset from base of record) (64 bytes):
0x0000 - 0x0003: Flags
  Bit 0: set = channel reversed
0x0004 - 0x0007: Trim value (-100 to +100, integer)
0x0008 - 0x000B: LeftDown travel volume (+10 to +120, integer)
0x000C - 0x000F: RightUp travel volume (+10 to +120, integer)
0x0010 - 0x003F: Reserved
*/

// maximum number of bytes that can be written at once
#define MAXWRITESIZE 32
// length of an address in bytes
#define ADDRESSLENGTH 2

// base address of global configuration
#define GLOBALCONFIGURATION_BASE 0x0000
// base address of aircraft records
#define AIRCRAFTRECORDS_BASE 0x0400
// gets the aircraft record base address from an aircraft number
#define AIRCRAFTRECORD_BASE(aircraft) AIRCRAFTRECORDS_BASE + (0x0400 * (aircraft - 1))
// gets a channel record base address from aircraft and channel numbers
#define CHANNELRECORD_BASE(aircraft, channel) (AIRCRAFTRECORD_BASE(aircraft)) + 0x0040 + (0x0040 * (channel - 1))

// address of last known throttle value
#define LASTKNOWNTHROTTLE (GLOBALCONFIGURATION_BASE + 0x0004)

// supported channel flag bits
#define FLAG_REVERSED 0

// the number of the I2C interface to use (0 - 2)
#define I2CINTERFACE 2

// I2C address of memory device
static const int EEPROMAddress = 0x50 << 1;

// initializes the configuration module
void Configuration_Init
  (
  void
  )
{
  // initialize I2C for EEPROM
#if I2CINTERFACE == 0
  I2C0Init();
#elif I2CINTERFACE == 1
  I2C1Init();
#else
  I2C2Init();
#endif

  // disable writing
  Core_ConfigurationWriteDisable(); 
}

// write data to memory
static void Configuration_Write
  (
  uint16_t Address,                                        // address to start writing to
  uint8_t *Data,                                           // data to write
  int Length                                               // number of bytes to write
  )
{
  int i;

  // enable writing
  Core_ConfigurationWriteEnable(); 

  // write
  do
  {
    I2CWriteLength[I2CINTERFACE] = Length + 1 + ADDRESSLENGTH;
    I2CReadLength[I2CINTERFACE] = 0;
    // device address
    I2CMasterBuffer[I2CINTERFACE][0] = EEPROMAddress;
    // address in big-endian format
    I2CMasterBuffer[I2CINTERFACE][1] = (Address >> 8) & 0xFF;
    I2CMasterBuffer[I2CINTERFACE][2] = Address & 0xFF;
    // data
    for (i = 0; i < Length; i++)
    {
      I2CMasterBuffer[I2CINTERFACE][1 + ADDRESSLENGTH + i] = Data[i];
    }
  }
  while (I2CEngine(I2CINTERFACE) == I2C_NACK_ON_ADDRESS);

  // disable writing
  Core_ConfigurationWriteDisable(); 
}

// read data from memory
static void Configuration_Read
  (
  uint16_t Address,                                        // address to start reading from
  uint8_t *Data,                                           // location to store data
  int Length                                               // number of bytes to read
  )
{
  int i;

  do
  {
    I2CWriteLength[I2CINTERFACE] = 3;
    I2CReadLength[I2CINTERFACE] = Length;
    I2CMasterBuffer[I2CINTERFACE][0] = EEPROMAddress;
    I2CMasterBuffer[I2CINTERFACE][1] = (Address >> 8) & 0xFF;
    I2CMasterBuffer[I2CINTERFACE][2] = Address & 0xFF;
    I2CMasterBuffer[I2CINTERFACE][3] = EEPROMAddress | RD_BIT;
  }
  while (I2CEngine(I2CINTERFACE) == I2C_NACK_ON_ADDRESS);

  for (i = 0; i < Length; i++)
  {
    Data[i] = I2CSlaveBuffer[I2CINTERFACE][i];
  }
}

// selects an aircraft to use
void Configuration_SetCurrentAircraft
  (
  uint32_t Aircraft                                        // number of the aircraft (1 - MAXAIRCRAFT)
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  Configuration_Write(GLOBALCONFIGURATION_BASE, (uint8_t *)&Aircraft, 4);
}

// gets the currently selected aircraft
uint32_t Configuration_GetCurrentAircraft
  (
  void
  )
{
  uint8_t Buffer[4];
    
  Configuration_Read(GLOBALCONFIGURATION_BASE, Buffer, 4);
    
  return *((uint32_t *)&Buffer);
}

// stores the current throttle value
void Configuration_SetCurrentThrottle
  (
  int32_t ThrottleValue                                    // throttle value (-100 to 100)
  )
{
  Configuration_Write(LASTKNOWNTHROTTLE, (uint8_t *)&ThrottleValue, 4);
}

// gets the last known throttle value
uint32_t Configuration_GetLastKnownThrottle
  (
  void
  )
{
  uint8_t Buffer[4];
    
  Configuration_Read(LASTKNOWNTHROTTLE, Buffer, 4);
    
  return *((uint32_t *)&Buffer);
}

// sets the number of channels for an aircraft
void Configuration_SetAircraftNumChannels
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t NumChannels                                     // number of channels
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  Configuration_Write(AIRCRAFTRECORD_BASE(Aircraft), (uint8_t *)&NumChannels, 4);
}

// sets the name of an aircraft
void Configuration_SetAircraftName
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  char *Name                                               // name of the aircraft (max 31 characters)
  )
{
  uint8_t Buffer[32];
  int i;

  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  // copy name
  for (i = 0; i < 31; i++)
  {
    Buffer[i] = Name[i];
    if (Buffer[i] == 0) break;
  }
  // guarantee termination
  Buffer[31] = 0;

  // write
  Configuration_Write(AIRCRAFTRECORD_BASE(Aircraft) + 0x0020, Buffer, 32);
}

// sets the direction of the channel (reversed or not reversed)
void Configuration_SetChannelDirection
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel,                                        // number of the channel (1 - MAXCHANNELS)
  int Reverse                                              // zero = not reversed, non-zero = reversed
  )
{
  uint32_t Flags;

  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  // read current flags
  Configuration_Read(CHANNELRECORD_BASE(Aircraft, Channel), (uint8_t *)&Flags, 4);

  // set or clear flag
  if (Reverse)
  {
    Flags |= (1UL << FLAG_REVERSED);
  }
  else
  {
    Flags &= ~(1UL << FLAG_REVERSED);
  }
  
  // write new flags
  Configuration_Write(CHANNELRECORD_BASE(Aircraft, Channel), (uint8_t *)&Flags, 4);
}  

// sets the trim of the channel
void Configuration_SetChannelTrim
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel,                                        // number of the channel (1 - MAXCHANNELS)
  int32_t Trim                                             // -100 to +100
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  if (Trim < -100) Trim = -100;
  if (Trim > 100) Trim = 100;

  Configuration_Write(CHANNELRECORD_BASE(Aircraft, Channel) + 0x0004, (uint8_t *)&Trim, 4);
}

// sets the left/down travel volume of the channel
void Configuration_SetChannelLeftDownTravelVolume
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel,                                        // number of the channel (1 - MAXCHANNELS)
  uint32_t TravelVolume                                    // +10 to +120
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  if (TravelVolume < 10) TravelVolume = 10;
  if (TravelVolume > 120) TravelVolume = 120;

  Configuration_Write(CHANNELRECORD_BASE(Aircraft, Channel) + 0x0008, (uint8_t *)&TravelVolume, 4);
}

// sets the right/up travel volume of the channel
void Configuration_SetChannelRightUpTravelVolume
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel,                                        // number of the channel (1 - MAXCHANNELS)
  uint32_t TravelVolume                                    // +10 to +120
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  if (TravelVolume < 10) TravelVolume = 10;
  if (TravelVolume > 120) TravelVolume = 120;

  Configuration_Write(CHANNELRECORD_BASE(Aircraft, Channel) + 0x000C, (uint8_t *)&TravelVolume, 4);
}

// sets the channel assignments
void Configuration_SetChannelAssignments
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t ThrottleChannel,                                // channel number for throttle (1 - MAXCHANNELS)
  uint32_t AileronsChannel,                                // channel number for ailerons (1 - MAXCHANNELS)
  uint32_t RudderChannel,                                  // channel number for rudder (1 - MAXCHANNELS)
  uint32_t ElevatorsChannel                                // channel number for elevators (1 - MAXCHANNELS)
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (ThrottleChannel < 1) ThrottleChannel = 1;
  if (ThrottleChannel > MAXCHANNELS) ThrottleChannel = MAXCHANNELS;

  if (AileronsChannel < 1) AileronsChannel = 1;
  if (AileronsChannel > MAXCHANNELS) AileronsChannel = MAXCHANNELS;

  if (RudderChannel < 1) RudderChannel = 1;
  if (RudderChannel > MAXCHANNELS) RudderChannel = MAXCHANNELS;

  if (ElevatorsChannel < 1) ElevatorsChannel = 1;
  if (ElevatorsChannel > MAXCHANNELS) ElevatorsChannel = MAXCHANNELS;

  Configuration_Write(AIRCRAFTRECORD_BASE(Aircraft) + 0x0240, (uint8_t *)&ThrottleChannel,  4);
  Configuration_Write(AIRCRAFTRECORD_BASE(Aircraft) + 0x0244, (uint8_t *)&AileronsChannel,  4);
  Configuration_Write(AIRCRAFTRECORD_BASE(Aircraft) + 0x0248, (uint8_t *)&RudderChannel,    4);
  Configuration_Write(AIRCRAFTRECORD_BASE(Aircraft) + 0x024C, (uint8_t *)&ElevatorsChannel, 4);
}

// gets the number of channels for an aircraft
uint32_t Configuration_GetAircraftNumChannels
  (
  uint32_t Aircraft                                        // number of the aircraft (1 - MAXAIRCRAFT)
  )
{
  uint8_t Buffer[4];

  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  Configuration_Read(AIRCRAFTRECORD_BASE(Aircraft), Buffer, 4);
    
  return *((uint32_t *)&Buffer);
}

// gets the name of an aircraft
void Configuration_GetAircraftName
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  char *Name                                               // location to store name of the aircraft (max 31 characters)
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  // read
  Configuration_Read(AIRCRAFTRECORD_BASE(Aircraft) + 0x0020, (uint8_t *)Name, 32);
  
  // guarantee termination
  Name[31] = 0;
}

// gets the direction of the channel (reversed or not reversed)
// returns zero if not reversed, returns 1 if reversed
int Configuration_GetChannelDirection
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel                                         // number of the channel (1 - MAXCHANNELS)
  )
{
  uint32_t Flags;

  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  // read current flags
  Configuration_Read(CHANNELRECORD_BASE(Aircraft, Channel), (uint8_t *)&Flags, 4);

  return (Flags >> FLAG_REVERSED) & 0x1;
}  

// gets the trim of the channel
// returns the trim -100 to +100
int32_t Configuration_GetChannelTrim
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel                                         // number of the channel (1 - MAXCHANNELS)
  )
{
  int32_t Trim;

  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  Configuration_Read(CHANNELRECORD_BASE(Aircraft, Channel) + 0x0004, (uint8_t *)&Trim, 4);

  if (Trim < -100) Trim = -100;
  if (Trim > 100) Trim = 100;

  return Trim;
}

// gets the left/down travel volume of the channel
uint32_t Configuration_GetChannelLeftDownTravelVolume
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel                                         // number of the channel (1 - MAXCHANNELS)
  )
{
  uint32_t TravelVolume;

  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  Configuration_Read(CHANNELRECORD_BASE(Aircraft, Channel) + 0x0008, (uint8_t *)&TravelVolume, 4);

  if (TravelVolume < 10) TravelVolume = 10;
  if (TravelVolume > 120) TravelVolume = 120;

  return TravelVolume;
}

// gets the right/up travel volume of the channel
uint32_t Configuration_GetChannelRightUpTravelVolume
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t Channel                                         // number of the channel (1 - MAXCHANNELS)
  )
{
  uint32_t TravelVolume;

  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  if (Channel < 1) Channel = 1;
  if (Channel > MAXCHANNELS) Channel = MAXCHANNELS;

  Configuration_Read(CHANNELRECORD_BASE(Aircraft, Channel) + 0x000C, (uint8_t *)&TravelVolume, 4);

  if (TravelVolume < 10) TravelVolume = 10;
  if (TravelVolume > 120) TravelVolume = 120;

  return TravelVolume;
}

// sets the channel assignments
void Configuration_GetChannelAssignments
  (
  uint32_t Aircraft,                                       // number of the aircraft (1 - MAXAIRCRAFT)
  uint32_t *ThrottleChannel,                               // filled with channel number for throttle
  uint32_t *AileronsChannel,                               // filled with channel number for ailerons
  uint32_t *RudderChannel,                                 // filled with channel number for rudder
  uint32_t *ElevatorsChannel                               // filled with channel number for elevators
  )
{
  if (Aircraft < 1) Aircraft = 1;
  if (Aircraft > MAXAIRCRAFT) Aircraft = MAXAIRCRAFT;

  Configuration_Read(AIRCRAFTRECORD_BASE(Aircraft) + 0x0240, (uint8_t *)ThrottleChannel,  4);
  Configuration_Read(AIRCRAFTRECORD_BASE(Aircraft) + 0x0244, (uint8_t *)AileronsChannel,  4);
  Configuration_Read(AIRCRAFTRECORD_BASE(Aircraft) + 0x0248, (uint8_t *)RudderChannel,    4);
  Configuration_Read(AIRCRAFTRECORD_BASE(Aircraft) + 0x024C, (uint8_t *)ElevatorsChannel, 4);
}
