// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// http://creativecommons.org/licenses/by-nc-sa/3.0/
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// $Id: main.c 402 2011-10-22 20:56:35Z andy $

#include "global.h"
#include "core.h"
#include "aircraft.h"
#include "configuration.h"
#include "commandparser.h"

#if USENUNCHUK
#include "nunchuk.h"
// time between reads of the nunchuk in milliseconds
#define NUNCHUKREADPERIOD 10
// time between reads of the nunchuk when trimming in milliseconds
#define NUNCHUKTRIMPERIOD 250
// minimum amount of motion in each direction to cause trimming values to be changed
// as a percentage when using nunchuk
#define LEFTDOWN_TRIMMIN -70
#define RIGHTUP_TRIMMIN 70
// amount as a percentage that the throttle trim is increased or decreased by
// when using nunchuk
#define THROTTLETRIMAMOUNT 10
// scaling factor for throttle to avoid having to point the nunchuk straight up for
// full throttle - makes usage more comfortable
#define THROTTLESCALING 1.4
#endif // USENUNCHUK

int main
  (
	void
	)
{
  uint32_t Throttle, Rudder, Ailerons, Elevator;
  uint32_t Aircraft;
  uint32_t NumberofChannels;
  uint32_t Channel;
  boolean_t WatchdogReset = False;
#if USENUNCHUK
  int16_t JoystickX;
  int16_t JoystickY;
  int16_t NunchukX;
  int16_t NunchukY;
  boolean_t CButton;
  boolean_t ZButton;
  boolean_t PreviousCButton = False;
  uint32_t NunchukReadTime;
  uint32_t TrimUpdateTime;
  int32_t NewTrim;
#endif // USENUNCHUK

  // determine if last reset was caused by the watchdog or not
  if (Core_IsLastResetFromWatchdog()) WatchdogReset = True;

  // initialize the core microcontroller module
  Core_Init();
  // initialize configuration module
  Configuration_Init();
  // initialize command parser module and start receiving commands
  CommandParser_Init();
#if USENUNCHUK
  // initialize nuchuck module
  Nunchuk_Init();
  // get next nunchuk read time
  NunchukReadTime = Core_GetTime() + NUNCHUKREADPERIOD;
  // get next nunchuk-based trim update time
  TrimUpdateTime = Core_GetTime() + NUNCHUKTRIMPERIOD;
#endif // USENUNCHUK

  // get current aircraft
  Aircraft = Configuration_GetCurrentAircraft();

  // get number of channels
  NumberofChannels = Configuration_GetAircraftNumChannels(Aircraft);

  // initialize aicraft module and start communicating with transmitter
	Aircraft_Init(NumberofChannels);

  // get channel assignments
  Configuration_GetChannelAssignments(Aircraft, &Throttle, &Ailerons, &Rudder, &Elevator);

  // keep going...
  Core_FeedWatchdog();

  // apply channel configurations
  for (Channel = 0; Channel < NumberofChannels; Channel++)
  {
    // direction
    Aircraft_SetChannelDirection((int)Channel, Configuration_GetChannelDirection(Aircraft, Channel));

    // trim
    Aircraft_SetTrim((int)Channel, Configuration_GetChannelTrim(Aircraft, Channel));

    // left/down travel volume
    Aircraft_SetTravelVolume((int)Channel, LeftDown, Configuration_GetChannelLeftDownTravelVolume(Aircraft, Channel));

    // right/up travel volume
    Aircraft_SetTravelVolume((int)Channel, RightUp, Configuration_GetChannelRightUpTravelVolume(Aircraft, Channel));
  }

  // if the reset was caused by the watchdog then restore the throttle
  // last used to keep aircraft flying, otherwise set engine to idle
  if (WatchdogReset)
  {
    // restore throttle
    Aircraft_SetStick(Throttle, Configuration_GetLastKnownThrottle());
  }
  else
  {
    // initialize throttle - engine idle
    Aircraft_SetStick(Throttle, -100);
    Configuration_SetCurrentThrottle(-100);
  }

  // turn status LED on to indicate initialization has completed
  Core_StatusLEDOn();

  while(1)
  {
    // keep executing
    Core_FeedWatchdog();

#if USENUNCHUK
    // if time to read nunchuk then read it now
    if (Core_IsTimeExpired(NunchukReadTime))
    {
      // get time of next nunchuk read
      NunchukReadTime = Core_GetTime() + NUNCHUKREADPERIOD;

      Nunchuk_Read(&JoystickX, &JoystickY, &NunchukX, &NunchukY, &CButton, &ZButton);

      // if c button pressed then in trimming mode
      if (CButton)
      {
        // if c button just pressed then reset trim timer
        if (!PreviousCButton)
        {
          TrimUpdateTime = 0;
          PreviousCButton = CButton;
        }

        // if time to update trim settings then do it now
        if (Core_IsTimeExpired(TrimUpdateTime))
        {
          if (NunchukX <= LEFTDOWN_TRIMMIN)
          {
            NewTrim = Aircraft_SubtractTrim(Ailerons);
            Configuration_SetChannelTrim(Aircraft, Ailerons, NewTrim);
          }
          else if (NunchukX >= RIGHTUP_TRIMMIN)
          {
            NewTrim = Aircraft_AddTrim(Ailerons);
            Configuration_SetChannelTrim(Aircraft, Ailerons, NewTrim);
          }

          if (JoystickX <= LEFTDOWN_TRIMMIN)
          {
            NewTrim = Aircraft_SubtractTrim(Rudder);
            Configuration_SetChannelTrim(Aircraft, Rudder, NewTrim);
          }
          else if (JoystickX >= RIGHTUP_TRIMMIN)
          {
            NewTrim = Aircraft_AddTrim(Rudder);
            Configuration_SetChannelTrim(Aircraft, Rudder, NewTrim);
          }

          if (JoystickY <= LEFTDOWN_TRIMMIN)
          {
            NewTrim = Aircraft_SubtractTrim(Elevator);
            Configuration_SetChannelTrim(Aircraft, Elevator, NewTrim);
          }
          else if (JoystickY >= RIGHTUP_TRIMMIN)
          {
            NewTrim = Aircraft_AddTrim(Elevator);
            Configuration_SetChannelTrim(Aircraft, Elevator, NewTrim);
          }

          // throttle trim is only adjustable if z button is pressed
          // to stop accidental changes to throttle
          if (ZButton)
          {
            if (NunchukY <= LEFTDOWN_TRIMMIN)
            {
              NewTrim = Aircraft_SubtractTrimByAmount(Throttle, THROTTLETRIMAMOUNT);
              Configuration_SetChannelTrim(Aircraft, Throttle, NewTrim);
            }
            else if (NunchukY >= RIGHTUP_TRIMMIN)
            {
              NewTrim = Aircraft_AddTrimByAmount(Throttle, THROTTLETRIMAMOUNT);
              Configuration_SetChannelTrim(Aircraft, Throttle, NewTrim);
            }
          }

          // another refresh in a while
          TrimUpdateTime = Core_GetTime() + NUNCHUKTRIMPERIOD;
        }
      }
      // c button not pressed so in normal mode
      else
      {
        // if z button is pressed then nunchuk y axis controls throttle
        if (ZButton)
        {
          Aircraft_SetStick(Throttle, (int)((double)(NunchukY + 100) * THROTTLESCALING) - 100);          
        }

        // use joystick for rudder and elevator
        Aircraft_SetStick(Rudder, JoystickX);
        Aircraft_SetStick(Elevator, JoystickY);

        // use nunchuk X axis for ailerons but with a large
        // "dead space" in the middle of the range to make it easy
        // to stop the roll
        if (NunchukX >= 50)
        {
          NunchukX = (NunchukX - 50) * 2;
        }
        else if (NunchukX <= -50)
        {
          NunchukX = (NunchukX + 50) * 2;
        }
        else
        {
          NunchukX = 0;
        }
        Aircraft_SetStick(Ailerons, NunchukX);

      }
    }
#endif // USENUNCHUK

    // parse the next serial command, if there is one
    CommandParser_Parse();
  }
}
