// nunchuk interface
// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// http://creativecommons.org/licenses/by-nc-sa/3.0/
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// $Id: nunchuk.c 393 2011-10-17 21:38:31Z andy $

// Raw values:
// X axis - roll left and right
//   Left = 140H, middle = 200H, right = 2C0H
// Y axis - pull back or push forwards
//   Level = 200H (512), middle = 1A0H (416), up = 140H (320)

#include "global.h"
#include "i2c.h"
#include "nunchuk.h"
#include "core.h"

#if USENUNCHUK

// X is XZERO when level
#define XZERO 0x200
//#define XMIN  0x160
//#define XMAX  0x2A0
#define XMIN  0x140
#define XMAX  0x2C0

// Y is YZERO when held at 45 degree angle
#define YZERO 0x188
#define YMIN  0x140
#define YMAX  0x1D0

// Z is zero when swicthing from right way up to upside down
#define ZZERO 0x200

// joystick X limits and center
#define JOYSTICK_XMIN  0x22
#define JOYSTICK_XMAX  0xE2
#define JOYSTICK_XZERO 0x82

// joystick Y limits and center
#define JOYSTICK_YMIN  0x22
#define JOYSTICK_YMAX  0xE2
#define JOYSTICK_YZERO 0x82

// number of accelerometer samples to average
#define MAXSAMPLES 8

typedef struct _average_t
{
  int16_t Samples[MAXSAMPLES];
  int NextSample;
  int16_t LastAverage;
} average_t;

static average_t XAxis;
static average_t YAxis;
static average_t ZAxis;

static int16_t LastX;
static int16_t LastY;

static void InitAxis(int16_t DefaultValue, average_t *Axis);

// initializes the nunchuk module and performs handshake with nunchuk
void Nunchuk_Init
  (
	void
	)
{
  I2C1Init();

  // init axes
  InitAxis(XZERO, &XAxis);
  InitAxis(YZERO, &YAxis);
  InitAxis(ZZERO, &ZAxis);

  // previous values
  LastX = XZERO;
  LastY = YZERO;

  // handshake
  do
  {
    I2CWriteLength[1] = 3;
    I2CReadLength[1] = 0;
    I2CMasterBuffer[1][0] = 0x52 << 1;
    I2CMasterBuffer[1][1] = 0xF0;
    I2CMasterBuffer[1][2] = 0x55;
  }
  while (I2CEngine(1) == I2C_NACK_ON_ADDRESS);

  // wait 1ms
  Core_Delay(1);

  do
  {
    I2CWriteLength[1] = 3;
    I2CReadLength[1] = 0;
    I2CMasterBuffer[1][0] = 0x52 << 1;
    I2CMasterBuffer[1][1] = 0xFB;
    I2CMasterBuffer[1][2] = 0x00;
  }
  while (I2CEngine(1) == I2C_NACK_ON_ADDRESS);

  // wait 1ms before doing anything else with nunchuk
  Core_Delay(1);
}

// gets the latest average value for an axis
static int16_t GetAverage
  (
  int16_t NewValue,                                        // new sample for axis
  average_t *Axis                                          // axis declaration
  )
{
  int16_t OldestValue;

  NewValue /= MAXSAMPLES;
  OldestValue = Axis->Samples[Axis->NextSample];
  Axis->Samples[Axis->NextSample] = NewValue;
  if (++(Axis->NextSample) == MAXSAMPLES) Axis->NextSample = 0;

  // calcluate average
  Axis->LastAverage = Axis->LastAverage - OldestValue + NewValue;
  
  return Axis->LastAverage;
}

// initializes an axis so that the first average result will
// be a default value
static void InitAxis
  (
  int16_t DefaultValue,                                    // default value for axis
  average_t *Axis                                          // axis declaration
  )
{
  int i;
  for (i = 0; i < MAXSAMPLES; i++) Axis->Samples[i] = DefaultValue / MAXSAMPLES;
  Axis->LastAverage = DefaultValue;
  Axis->NextSample = 0;
}

// read current values from nunchuk
// takes about 1.3ms to execute
void Nunchuk_Read
  (
  int16_t *JoystickX,                                      // -100 to +100
  int16_t *JoystickY,                                      // -100 to +100
  int16_t *NunchukX,                                       // -100 (left) to +100 (right)
  int16_t *NunchukY,                                       // -100 (level) to +100 (pointing up)
  boolean_t *CButton,                                      // true if pressed
  boolean_t *ZButton                                       // true if pressed
  )
{
  float X, Y;
  int16_t NunchukZ;

  I2CWriteLength[1] = 2;
  I2CReadLength[1] = 0;
  I2CMasterBuffer[1][0] = 0x52 << 1;
  I2CMasterBuffer[1][1] = 0x00;
  I2CEngine(1);

  // wait 1ms
  Core_Delay(1);

  I2CWriteLength[1] = 0;
  I2CReadLength[1] = 6;
  I2CMasterBuffer[1][0] = (0x52 << 1) | RD_BIT;
  I2CEngine(1);

  // read joystick
  X = I2CSlaveBuffer[1][0];
  Y = I2CSlaveBuffer[1][1];

  // scale joystick x to -100 -> +100
  X -= JOYSTICK_XZERO;
  X *= 200.0 / (JOYSTICK_XMAX - JOYSTICK_XMIN);
  if (X < -100.0) X = -100.0; 
  if (X > 100.0) X = 100.0;

  // scale joystick y to -100 -> +100
  Y -= JOYSTICK_YZERO;
  Y *= 200.0 / (JOYSTICK_YMAX - JOYSTICK_YMIN);
  if (Y < -100.0) Y = -100.0; 
  if (Y > 100.0) Y = 100.0;

  // new joystick values
  *JoystickX = X;
  *JoystickY = Y;

  // get accelerometer averages
  X = (float)GetAverage(((int16_t)(I2CSlaveBuffer[1][2]) << 2) | ((I2CSlaveBuffer[1][5] >> 2) & 0x3), &XAxis);
  Y = (float)GetAverage(((int16_t)(I2CSlaveBuffer[1][3]) << 2) | ((I2CSlaveBuffer[1][5] >> 4) & 0x3), &YAxis);
  NunchukZ = GetAverage(((int16_t)(I2CSlaveBuffer[1][4]) << 2) | ((I2CSlaveBuffer[1][5] >> 6) & 0x3), &ZAxis);

  // scale X to -100 -> +100
  X -= XZERO;
  X *= 200.0 / (XMAX - XMIN);
  if (X < -100.0) X = -100.0; 
  if (X > 100.0) X = 100.0;

  // scale Y to -100 -> +100
  Y -= YZERO;
  Y = -Y;
  Y *= 200.0 / (YMAX - YMIN);
  if (Y < -100.0) Y = -100.0; 
  if (Y > 100.0) Y = 100.0;

  // if nunchuk upside down then use the last valid values
  if (NunchukZ < ZZERO)
  {
    *NunchukX = LastX;
    *NunchukY = LastY;
  }
  // nunchuk right way up so use new values
  else
  {
    // remember values
    LastX = (int16_t)X;
    LastY = (int16_t)Y;

    // return values
    *NunchukX = LastX;
    *NunchukY = LastY;
  }

  // read Z button and invert (to give pressed = true)
  if (I2CSlaveBuffer[1][5] & 0x1)
    *ZButton = False;
  else
    *ZButton = True;

  // read C button and invert (to give pressed = true)
  if ((I2CSlaveBuffer[1][5] >> 1) & 0x1)
    *CButton = False;
  else
    *CButton = True;
}

#endif // USENUNCHUK
