// core microcontroller specific functionality
// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// http://creativecommons.org/licenses/by-nc-sa/3.0/
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// $Id: core_lpc17xx.c 393 2011-10-17 21:38:31Z andy $

#include "global.h"
#include "lpc17xx.h"
#include "core.h"

// incremented every millisecond - keeps track of time
static volatile uint32_t TickCounter;

void Core_Init
  (
	void
	)
{
  // CCLK = 96MHz for 12MHz crystal
  SystemInit();

  // P1.18 (LED1/Status)
  // P1.18 is gpio and output
  LPC_PINCON->PINSEL3 &= ~(3UL << 4);
  LPC_GPIO1->FIODIR |= (1UL << 18);
  // led off
  LPC_GPIO1->FIOCLR = (1UL << 18);

  // configure write enable signal
  // output on P2.0 (mbed P26)
  // P2.0 is gpio
  LPC_PINCON->PINSEL4 &= ~(3UL << 0);
  // P2.0 is output
  LPC_GPIO2->FIODIR |= (1UL << 0);

  // clear counter
  TickCounter = 0;
  // generates 1ms interrupts
  // pclk = cclk / 4 = 24MHz
  // 24MHz / 24000 = 1kHz
  LPC_RIT->RICOMPVAL = 24000 - 1;
  // match all values
  LPC_RIT->RIMASK = 0x00000000;
  // set up RIT interrupt
  // must be highest priority after ppm module
  NVIC_SetPriority(RIT_IRQn, 2);
  NVIC_EnableIRQ(RIT_IRQn);
  // enable timer, clear value on interrupt
  LPC_RIT->RICTRL = 0x0000000E;
  
  // init watchdog to use IRC oscillator and cause a reset on timeout
  // resets after 100ms if not fed
  LPC_WDT->WDCLKSEL = 0x80000000;
  // IRC (4MHz) / 4 / 10,000 = 100ms (10Hz)
  LPC_WDT->WDTC = 10000;
  LPC_WDT->WDMOD = 0x00000003;
  // enable watchdog
  __disable_irq();
  LPC_WDT->WDFEED = 0xAA;
  LPC_WDT->WDFEED = 0x55;
  __enable_irq();
}

// gets the current timestamp in milliseconds
uint32_t Core_GetTime
  (
  void
  )
{
  return TickCounter;
}

// checks if a timestamp is in the past
// returns true if in the past
boolean_t Core_IsTimeExpired
  (
  uint32_t Timestamp                                       // timestamp to check
  )
{
  uint32_t TimeNow;

  TimeNow = Core_GetTime();
  if (TimeNow >= Timestamp)
  {
    if ((TimeNow - Timestamp) < 0x80000000)
      return True;
    else
      return False;
  }
  else
  {
    if ((Timestamp - TimeNow) >= 0x80000000)
      return True;
    else
      return False;
  }
}

// returns after waiting a specific number of milliseconds
void Core_Delay
  (
  uint32_t Milliseconds                                    // time to wait
  )
{
  uint32_t EndTime = Core_GetTime() + Milliseconds;
  while (!Core_IsTimeExpired(EndTime));
}

// RIT interrupt handler
// called once per millisecond, keeps track of time
void RIT_IRQHandler()
{
  __disable_irq();

  // increment counter
  TickCounter++;

  __enable_irq();

  // clear interrupt
  LPC_RIT->RICTRL |= 0x1;
}

// turns the status led on
void Core_StatusLEDOn
  (
  void
  )
{
  LPC_GPIO1->FIOSET = (1UL << 18);
}

// turns the status led off
void Core_StatusLEDOff
  (
  void
  )
{
  LPC_GPIO1->FIOCLR = (1UL << 18);
}

// enable write access to configuration storage
void Core_ConfigurationWriteEnable
  (
  void
  )
{
  LPC_GPIO2->FIOCLR = (1UL << 0);
}

// disable write access to configuration storage
void Core_ConfigurationWriteDisable
  (
  void
  )
{
  LPC_GPIO2->FIOSET = (1UL << 0);
}

// feeds the watchdog to stop the processor from resetting
void Core_FeedWatchdog
  (
  void
  )
{
  LPC_WDT->WDFEED = 0xAA;
  LPC_WDT->WDFEED = 0x55;
}

// checks if the last processor reset was caused by the watchdog or not
// returns true if watchdog reset processor
boolean_t Core_IsLastResetFromWatchdog
  (
  void
  )
{
  if (((LPC_WDT->WDMOD >> 2) & 0x1) == 0x1)
  {
    // clear flag
    LPC_WDT->WDMOD &= ~(1UL << 2);
    return True;
  }

  return False;
}

// resets the microcontroller after a delay
void Core_Reset
  (
  void
  )
{
  Core_FeedWatchdog();

  // wait for watchdog to timeout and generate a reset
  while(1);
}
