// ppm module
// generates a six channel ppm waveform for feeding into a futaba rc transmitter
// for nxp lpc17xx ARM Cortex-M3 microcontroller
// assumes cpu clock is 92MHz
// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// http://creativecommons.org/licenses/by-nc-sa/3.0/
// $Id: ppm_lpc17xx.c 393 2011-10-17 21:38:31Z andy $

#include "lpc17xx.h"
#include "config.h"
#include "ppm.h"
#include "configuration.h"

// length of a frame of ppm data in milliseconds
// must be more than ((LOWPULSE + CHANNELMAX) * MAXCHANNELS) + (LOWPULSE * 2)
#define FRAMELENGTH 18
// length of low pulse before each channel high pulse in microseconds
#define LOWPULSE 400
// absolute maximum pulse width in microseconds
#define CHANNELMAX 2000
// absolute minimum pulse width in microseconds
#define CHANNELMIN 450
// max number of edges in a ppm frame
#define MAXEDGES ((MAXCHANNELS * 2) + 2)
// high time at the start of each ppm frame in microseconds
#define PREFRAMEPERIOD 100

// total number of tasks that can access the shadow registers (max eight)
#define NUMTASKS 2
// the tasks that can access the shadow registers
#define TASK_PPM 0
#define TASK_APPLICATION 1

// macros for setting state of output pin
#define PINHIGH   (LPC_GPIO2->FIOSET = (1UL << 5))
#define PINLOW    (LPC_GPIO2->FIOCLR = (1UL << 5))
#define PINSTATE ((LPC_GPIO2->FIOPIN >> 5) & 1)

// list of times between edges in microseconds
static unsigned long EdgeTimes[MAXEDGES];
// current position in edge times list
static int EdgeIndex = 0;

// the number of channels actually used
static int NumChannels;
// the number of waveform edges actually used
static int NumEdges;

// timer 0 interrupt
// generates a set of pulses, one for each channel, on the PPM signal and then stops
// walks through the edgetimes table to determine pulse widths
void TIMER0_IRQHandler()
{
    // MR0 match
    if (LPC_TIM0->IR & 1)
    {
        // toggle output
        if (PINSTATE)
          PINLOW;
        else
          PINHIGH;

        // move to next edge, if reached end then stop
        if (++EdgeIndex == NumEdges)
        {
            // stop and reset
            LPC_TIM0->TCR = 2;
        }
        else
        {
            // set up next edge
            LPC_TIM0->MR0 = EdgeTimes[EdgeIndex];
        }

        // clear interrupt
        LPC_TIM0->IR |= 1;
    }
}

// timer 1 interrupt
// gets new channel values and then triggers the generation of a new
// PPM frame. called every FRAMELENGTH milliseconds
void TIMER1_IRQHandler()
{
    // MR0 match
    if (LPC_TIM1->IR & 1)
    {
#if (INVERTPPMSIGNAL == 0)
        // signal is initially high
        PINHIGH;
#else
        // signal is initially low
        PINLOW;
#endif

        // set up first edge
        EdgeIndex = 0;
        LPC_TIM0->MR0 = EdgeTimes[EdgeIndex];

        // enable timer 0 to generate edges
        LPC_TIM0->TCR = 1;

        // clear interrupt
        LPC_TIM1->IR |= 1;
    }
}

// initializes the ppm module
// note: uses timers 0 and 1
void PPM_Init
    (
    unsigned int InitialValue,                             // initial value to use for all channels in microseconds
    int Channels                                           // number of channels
    ) 
{
    int Channel;

    // remember the number of channels and waveform edges
    NumChannels = Channels;
    NumEdges = (Channels * 2) + 2;

    // enable power to gpio
    LPC_SC->PCONP |= (1UL << 15);
    // output on P2.5 (mbed P21)
    // P2.5 is gpio
    LPC_PINCON->PINSEL4 &= ~(3UL << 10);
    // output
    LPC_GPIO2->FIODIR |= (1UL << 5);

    // enable power to timer 0
    LPC_SC->PCONP |= (1UL << 1);
    // timer 0 clock = cclk/4
    LPC_SC->PCLKSEL0 &= ~(3UL << 2);
    // set up timer 0 interrupt
    // must be highest priority
    NVIC_SetPriority(TIMER0_IRQn, 0);
    NVIC_EnableIRQ(TIMER0_IRQn);
    // pclk = 24MHz, divide down to 1MHz
    LPC_TIM0->PR = 24 - 1;
    // interrupt and reset on match 0
    LPC_TIM0->MCR = 3;
    // timer mode
    LPC_TIM0->CTCR = 0;

    // enable power to timer 1
    LPC_SC->PCONP |= (1UL << 2);
    // timer 1 clock = cclk/4
    LPC_SC->PCLKSEL0 &= ~(3UL << 4);
    // set up timer 1 interrupt
    // must be second-highest priority
    NVIC_SetPriority(TIMER1_IRQn, 1);
    NVIC_EnableIRQ(TIMER1_IRQn);
    // pclk = 24MHz, divide down to 1kHz
    LPC_TIM1->PR = 24000 - 1;
    // match register to generate an interrupt every FRAMELENGTH ms
    LPC_TIM1->MR0 = FRAMELENGTH - 1;
    // interrupt and reset on match 0
    LPC_TIM1->MCR = 3;
    // timer mode
    LPC_TIM1->CTCR = 0;

    // initialize edge times table
    // setup period
    EdgeTimes[0] = PREFRAMEPERIOD - 1;
    for (Channel = 0; Channel < NumChannels; Channel++)
    {
        // sync pulse
        EdgeTimes[(Channel << 1) + 1] = LOWPULSE - 1;
        // value pulse
        EdgeTimes[(Channel << 1) + 2] = InitialValue - 1;
    }
    // final sync pulse
    EdgeTimes[(NumChannels << 1) + 1] = LOWPULSE - 1;

#if (INVERTPPMSIGNAL == 0)
    // initial output state is high
    PINHIGH;
#else
    // initial output state is low
    PINLOW;
#endif
}

// starts the ppm waveform generation
void PPM_Start(void)
{
    // start PPM frame timer
    LPC_TIM1->TCR = 1;
}

// stops the ppm waveform generation
void PPM_Stop(void)
{
    // stop PPM frame timer
    LPC_TIM1->TCR = 2;
}

// sets the value of a channel in microseconds
void PPM_SetChannelValue
    (
    unsigned int Channel,                                  // number of channel (0 to NumChannels - 1)
    unsigned int Value                                     // new channel value in microseconds
    )
{
    int Index;

    // ignore unsupported channels
    if (Channel >= NumChannels) return;

    // make sure the value is with the allowed range
    if (Value > CHANNELMAX) Value = CHANNELMAX;
    if (Value < CHANNELMIN) Value = CHANNELMIN;

    // get index into edge times table
    Index = (Channel << 1) + 2;

    __disable_irq();

    // use value
    EdgeTimes[Index] = Value - 1;

    __enable_irq();
}

// gets the current value for a channel
// must not be called while PPM_SetChannel is being executed
// returns the current channel value in microseconds
unsigned int PPM_GetChannelValue
    (
    unsigned int Channel                                   // number of channel (0 to NumChannels - 1)
    )
{
    // get index into edge times table
    int Index = (Channel << 1) + 2;

    // ignore unsupported channels
    if (Channel >= NumChannels) return 0;
  
    // return current value
    return EdgeTimes[Index] + 1;
}
