// serial interface
// receives command terminated with CR + LF and sends responses
// terminated with CR + LF
// uses 57600, 8N1
// OpenRadioControl
// (C) BritishIdeas.com 2011 andy@britishideas.com
// This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
// http://creativecommons.org/licenses/by-nc-sa/3.0/
// THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
// OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
// BRITISHIDEAS SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
// $Id: serial_lpc17xx.c 396 2011-10-18 21:51:01Z andy $

#include "lpc17xx.h"
#include "serial.h"
#include "core.h"
#include "library.h"

// maximum time between receiving characters before
// giving up in milliseconds
#define MAXCHARACTERDELAY 2000

// control characters
#define NUL 0x00
#define LF  0x0A
#define CR  0x0D

// interrupt sources
#define RECEIVE_DATA_AVAILABLE (2 << 1)

// receive handling
static char ReceiveBuffer[SERIAL_MAXCOMMANDLENGTH];
static int ReceiveBufferIndex;
static uint32_t LastCharacterReceiveTime;

// flag to indicate if a command is waiting to be read
static int CommandReady;

// sends a response back
void Serial_SendResponse
  (
  response_t Response,                                     // response to send back
  boolean_t Wait                                           // false to return immediately, true to wait for transmit to complete
  )
{
  switch (Response)
  {
    case Ok:
      Serial_SendString("OK");
      break;

    default:
    case Error:
      Serial_SendString("ERROR");
      break;
  }

  // if requested, wait for transmit to complete
  if (Wait)
  {
    Serial_WaitForTransmitComplete();
  }
}

// UART0 interrupt handler
// processed received bytes
void UART0_IRQHandler()
{
  uint32_t InterruptSource = LPC_UART0->IIR;

  if (InterruptSource & RECEIVE_DATA_AVAILABLE)
  {
    // data arriving but the previous command was not
    // processed, so now we have a command overflow
    // clear receive buffer
    if (CommandReady)
    {
      ReceiveBufferIndex = 0;
    }

    // receiving data so no command ready to be processed
    CommandReady = 0;

    // if it has been too long since the last character was received
    // then clear receive buffer and start again
    if (Core_IsTimeExpired(LastCharacterReceiveTime + MAXCHARACTERDELAY))
    {
      ReceiveBufferIndex = 0;
    }

    // get time that the character was received
    LastCharacterReceiveTime = Core_GetTime();

    // store character in receive buffer
    ReceiveBuffer[ReceiveBufferIndex++] = LPC_UART0->RBR;

    // if more than two character received then check for end of command sequence
    if (ReceiveBufferIndex >= 2)
    {
      // if CR + LF received then we have a command
      if (ReceiveBuffer[ReceiveBufferIndex - 1] == LF &&
          ReceiveBuffer[ReceiveBufferIndex - 2] == CR)
      {
        // command is ready to be processed
        CommandReady = 1;
      }
    }

    // wrap on buffer overflow
    if (ReceiveBufferIndex == SERIAL_MAXCOMMANDLENGTH) ReceiveBufferIndex = 0;
  }
}

// gets the latest received command
// returns length of command received, zero if no command received
int32_t Serial_GetCommand
  (
  char *Command                                            // location to store received command
  )
{
  int i;
  int32_t Length;

  // if no command waiting to be read then nothing to do
  if (!CommandReady) return 0;

  // disable uart zero interrupt
  NVIC_DisableIRQ(UART0_IRQn);

  // copy command replacing CR+LF with NULL
  for (i = 0; i < ReceiveBufferIndex - 2; i++) Command[i] = ReceiveBuffer[i];
  Command[ReceiveBufferIndex - 2] = NUL;

  // store command length
  Length = ReceiveBufferIndex - 2;

  // no longer a command waiting to be read
  CommandReady = 0;

  // clear receive buffer
  ReceiveBufferIndex = 0;

  // enable uart zero interrupt
  NVIC_EnableIRQ(UART0_IRQn);

  return Length;
}

// initializes the serial module
void Serial_Init
  (
	void
	)
{
  ReceiveBufferIndex = 0;
  LastCharacterReceiveTime = 0;

  // UART 0 pclk = cclk/4 = 24MHz
  LPC_SC->PCLKSEL0 &= ~(3UL << 6);
  // 8 bits, no parity, 1 stop bit, no flow control, DLAB = 1
  LPC_UART0->LCR = 0x00000083;
  // for 57600 we need a clock rate of 57600 x 16 = 921600
  // 24MHz / 921.600kHz = 26 which gives 57692 baud
  LPC_UART0->DLM = 0;
  LPC_UART0->DLL = 26;
  // no fractional baud rate generation
  LPC_UART0->FDR = 0x00000010;
  // DLAB = 0
  LPC_UART0->LCR = 0x00000003;
  // enable FIFOs, generate receive interrupt for each character received
  LPC_UART0->FCR = 0x00000007;
  // select P0.2 and P0.3 for UART0 pins
  LPC_PINCON->PINSEL0 &= ~(0xFUL << 4);
  LPC_PINCON->PINSEL0 |= (0x5UL << 4);
  // set P0.2 and P0.3 to no pull ups or pull downs
  LPC_PINCON->PINMODE0 &= ~(0xFUL << 4);
  LPC_PINCON->PINMODE0 |= (0xAUL << 4);
  // enable receive interrupt
  LPC_UART0->IER = 0x00000001;
  // set up UART0 interrupt
  // lowest priority
  NVIC_SetPriority(UART0_IRQn, 10);
  NVIC_EnableIRQ(UART0_IRQn);
}

// transmits an ASCII character
void Serial_Send
  (
  uint8_t Value
  )
{
  // wait for THR empty
  while(!(LPC_UART0->LSR & (1 << 5)));
  // write
  LPC_UART0->THR = Value;
}

// waits until any pending characters have been transmitted
// warning: a watchdog reset may occur before this function completes
void Serial_WaitForTransmitComplete
  (
  void
  )
{
  while(!(LPC_UART0->LSR & (1 << 6)));
}

// transmits a string and appends line termination
void Serial_SendString
  (
  char *String                                             // string to transmit
  )
{
  int32_t i;

  for (i = 0; i < strlen(String); i++) Serial_Send(String[i]);
  Serial_Send(CR);
  Serial_Send(LF);
}

/*// transmits a hex byte as ASCII
void Serial_SendHex8
  (
  uint8_t Value
  )
{
  int Nibble = (Value >> 4) & 0xF;
  if (Nibble < 10)
    Serial_Transmit(Nibble + 0x30);
  else
    Serial_Transmit(Nibble - 10 + 0x41);

  Nibble = Value & 0xF;
  if (Nibble < 10)
    Serial_Transmit(Nibble + 0x30);
  else
    Serial_Transmit(Nibble - 10 + 0x41);
}*/
